/* =========================================================================
 * FILE     : vdocapture.c
 * ---------:---------------------------------------------------------------
 * By       : Bo-kyun Na (bkna@npac.syr.edu), IntellActive Research Inc.
 *          : 1999.10.4.
 * Revision : 
 * ========================================================================= */

#include "vdo.h"


/* *************************************************************************  
 * Function    : printctrl(struct v4l2_queryctrl *qc)
 * ------------:------------------------------------------------------------
 * Inputs      : struct v4l2_queryctrl *qc
 * Outputs     : void
 * Description :
 * ************************************************************************* */
void printctrl(struct v4l2_queryctrl *qc)
{	
  printf("  %-12s %-6s %8d %8d %7d %8d  %s\n", qc->name,
	 (qc->type == V4L2_CTRL_TYPE_MENU) ? "(menu)" : 
	 (qc->type == V4L2_CTRL_TYPE_BOOLEAN) ? "(bool)" : "(int)",
	 qc->minimum, qc->maximum, qc->step, qc->default_value, qc->group);
}


/* *************************************************************************  
 * Function    : BV_init_vdo_capture ()
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : void
 * Description :
 * ************************************************************************* */
int BV_vdo_init_capture (char *dev_vdo, VIMAGE *vimage, V4L2 *v4l2)
{
  int i;
  int err;
  int input;
  struct v4l2_queryctrl qc;


  vimage->vid = open(dev_vdo, O_RDONLY);
  if (vimage->vid < 0) {
    printf("No video device \"%s\"\n", dev_vdo);
    exit (0);
  }

  err = ioctl(vimage->vid, VIDIOC_QUERYCAP, &v4l2->cap);
  if (err) {
    printf("QUERYCAP returned error %d\n", errno);
    return 1;
  }

  if (verbose) {
    printf("Device:       %s\n", v4l2->cap.name);
    printf("Capabilities:");
    if (v4l2->cap.flags & V4L2_FLAG_READ)       printf(" capture via read();");
    if (v4l2->cap.flags & V4L2_FLAG_STREAMING)  printf(" streaming capture;");
    if (v4l2->cap.flags & V4L2_FLAG_PREVIEW)    printf(" automatic preview;");
    if (v4l2->cap.flags & V4L2_FLAG_SELECT)     printf(" select() call;");
    if (v4l2->cap.flags & V4L2_FLAG_TUNER)      printf(" tuner;");
    if (v4l2->cap.flags & V4L2_FLAG_MONOCHROME) printf(" monochrome only;");
    printf("\n");
    if (v4l2->cap.type != V4L2_TYPE_CAPTURE) {
      printf("Not a capture device.\n");
      return 1;
    }
  }

  if (verbose)
    printf("Video inputs on this device:\n");
  for (i = 0, err = 0; err == 0; ++i) {	
    struct v4l2_input inp;
    inp.index = i;
    err = ioctl(vimage->vid, VIDIOC_ENUMINPUT, &inp);
    if (!err && verbose)
      printf("  %d: \"%s\" is a%s, %s audio\n", i, inp.name,
	    (inp.type == V4L2_INPUT_TYPE_TUNER) ? "TV tuner" : "n analog input",
	     (inp.capability & V4L2_INPUT_CAP_AUDIO) ? "has" : "no");
  }

  input = VDO_INPUT;
  err = ioctl(vimage->vid, VIDIOC_S_INPUT, &input);
  if (err)
    printf("S_INPUT returned error %d\n",errno);

  if (verbose)
    printf("Video standards supported:");
  for (i = 0, err = 0; err == 0; ++i) {
    struct v4l2_enumstd	estd;
    estd.index = i;
    err = ioctl(vimage->vid, VIDIOC_ENUMSTD, &estd);
    if (!err && verbose) {
      printf("%c %s", i?',':' ', estd.std.name);
    }
  }
  if (verbose) 
    printf("\nVideo capture image formats supported:\n");
  for (i = 0, err = 0; err == 0; ++i) {
    struct v4l2_fmtdesc	fmtd;

    fmtd.index = i;
    err = ioctl(vimage->vid, VIDIOC_ENUM_PIXFMT, &fmtd);
    if (!err && verbose)
      printf("  %d: %s  (%s)\n", i, fmtd.description,
	     (fmtd.flags & V4L2_FMT_FLAG_COMPRESSED) ?
	     "compressed" : "uncompressed");
  }

  if (verbose) {
    printf("Controls supported:\n");
    printf("  Label        type    Minimum  Maximum    Step  Default Category\n");
  }
  qc.id = V4L2_CID_BRIGHTNESS;
  if ((ioctl(vimage->vid, VIDIOC_QUERYCTRL, &qc) == 0) && verbose) 
    printctrl(&qc);
  qc.id = V4L2_CID_CONTRAST;
  if ((ioctl(vimage->vid, VIDIOC_QUERYCTRL, &qc) == 0) && verbose)
    printctrl(&qc);
  qc.id = V4L2_CID_SATURATION;
  if ((ioctl(vimage->vid, VIDIOC_QUERYCTRL, &qc) == 0) && verbose)
    printctrl(&qc);
  qc.id = V4L2_CID_HUE;
  if ((ioctl(vimage->vid, VIDIOC_QUERYCTRL, &qc) == 0) && verbose)
    printctrl(&qc);

  if (!(v4l2->cap.flags & V4L2_FLAG_READ)) {
    printf("Device does not support the read() call.\n");
    return 1;
  }

  v4l2->parm.type = V4L2_BUF_TYPE_CAPTURE;
  ioctl(vimage->vid, VIDIOC_G_PARM, &v4l2->parm);

  /* set proper framerate */
  v4l2->parm.parm.capture.capturemode |= V4L2_CAP_TIMEPERFRAME;
  v4l2->parm.parm.capture.timeperframe = VDO_FRAMERATE;

  err = ioctl(vimage->vid, VIDIOC_S_PARM, &v4l2->parm);
  if (err)
    printf("S_PARM returned error %d\n",errno);

  v4l2->fmt.type = V4L2_BUF_TYPE_CAPTURE;
  
  return 0;
}



/* *************************************************************************
 * Function    : 
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
void BV_vdo_capture (unsigned char *capimg, VIMAGE *vimage, V4L2 *v4l2)
{

  read(vimage->vid, capimg, v4l2->fmt.fmt.pix.sizeimage);
}


/* *************************************************************************
 * Function    : 
 * ------------:------------------------------------------------------------
 * Inputs      : 
 * Outputs     : 
 * Description : 
 * ************************************************************************* */
