/* *************************************************************************
 * libMTodo.c
 *  > Audio related functions
 *  > programmer: Bo-kyun Na.    1999.02.16     all rights reserved. 
 * ************************************************************************* */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <poll.h>
#include <stropts.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/filio.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <sys/audioio.h>

#include <multimedia/libaudio.h>
#include <multimedia/audio_device.h>
#include <multimedia/audio_encode.h>

#include "oti_libaudio.h"
#include "oti_audio_device.h"
#include "params.h"
/* #include "odo.h" */
#include "network.h"
#include "transceive.h"

#define BUFFER_SIZE (1024*4)

int Audio_fd    = -1;
int Audioctl_fd = -1;
int input_fd, output_fd;

char aud_data[BUFFER_SIZE];
char DEFAULT_DEVICE[] = "/dev/o1k0";

int format   = 0;
int rate     = 8000;
int channels = 1;
unsigned int port = AUDIO_LINE_IN; 
unsigned int buffer_size = 1000;
double  igain = 0.125;
double  ogain = 1.0;  
double  mgain = 0.0;  /* looping audio: want monitor gain at 0 */

extern pthread_mutex_t utex_lock;
extern int count;
extern sem_t VS, VR, AS, AR;
extern void thrd_stop (sem_t *);

void *AudRecThread (void *);
void *AudPlayThread (void *);

/* *************************************************************************
 *  Function:	BV_stop ()
 *  Description: Stops looping by pausing the audio device and then 
 *               flushing it.
 *  Parameters: none
 *  Returns: void
 * ************************************************************************* */
void BV_stop()
{
  oti_audio_pause(Audio_fd);   /* Pause the device */
  oti_audio_flush(Audio_fd);   /* Flush the device */
}


/* *************************************************************************
 *  Function:	BV_exit
 *  Description:
 *	Hanldes signals which would cause process to exit (like Ctl-C).
 *      Stops looping and closes the various file descriptors. This
 *      routine can be called directly.
 *  Parameters:
 *	int signo: signal number
 *  Returns: void
 * ************************************************************************* */
void BV_exit(int signo)
{
  BV_stop();
  
  if (Audio_fd != -1) {
    oti_close(Audio_fd);
    Audio_fd = -1;
  }
  
  if (Audioctl_fd != -1) {
    oti_close(Audioctl_fd);
    Audioctl_fd = -1;
  }

  exit(0);
}


/* *************************************************************************
 *  Function:	BV_poll_handler
 *  Description:
 *	Hanldes SIGPOLL signal. A SIGPOLL is generated by the audio device
 *	when input and output ready conditions change. This is used to play
 *      or record audio data in an async, efficient manner.
 *  Parameters:
 *	int signo: signal number
 *  Returns: void
 * ************************************************************************* */
void BV_poll_handler (int signo)
{
  /* must re-signal immediately in case another signal occurs */
  signal(signo, BV_poll_handler);
}


/* *************************************************************************
 *  Function:	get_audioctl_name
 *
 *  Description:
 *	Given an audio device name, determine the control device name
 *	
 *  Parameters:
 *	char* adevice_ctl_name_ptr: ptr to place control device name into
 *      char* adevice_name_ptr:     ptr to audio device name
 *	
 *  Returns: void
 *	
 *  Side Effects:
 *	Modifies buffer pointed to by adevice_ctl_name_ptr. 
 *	
 *  Notes:
 *	adevice_ctl_name_ptr must point to a reasonable sized buffer
 * ************************************************************************* */
void get_audioctl_name (char* adevice_ctl_name_ptr, char* adevice_name_ptr)
{
  int i;
  char  c = '\0';
  char *s;

  /*
   *    Given         Produce      Example:
   *   xxxxaudio -> xxxxxaudioctl  /dev/audio -> /dev/audioctl
   *   xxxxo1kN  -> xxxxo1kctlN    /dev/o1k0  -> /dev/o1kctl0
   */

  strcpy(adevice_ctl_name_ptr, adevice_name_ptr);
  i = strlen(adevice_name_ptr);
  
  if (strstr(adevice_name_ptr,"o1k") != NULL){
    s = &adevice_ctl_name_ptr[i-1];
    c = *s;      
  } else {
    s = &adevice_ctl_name_ptr[i];
  }

  *s++ = 'c';
  *s++ = 't';
  *s++ = 'l';
  *s++ = c;
  *s++ = '\0';

}


/* *************************************************************************
 * BV_send_AudData ()
 *
 * ************************************************************************* */
void BV_send_AudData (char *data, int nbytes, TransceiveState *tstate, Parameters *param)
{
  PacketType packet;
  static int frame = 0;
  int count = 0;
  
  packet.header.frame = frame++;
  packet.header.session = tstate->session | (AUDIO_PACKET << 28);

  if (nbytes > PACKETDATASIZE * MAXPACKETS) {
    if (verbose)
      fprintf(stderr, "Not sending frame %d due to size (%d) > maximum\n",
              frame, nbytes);
  }
  else {
    while (nbytes > 0) {
      int curbytes = 0;
          
      if (nbytes <= PACKETDATASIZE) {
        curbytes = nbytes;
          packet.header.flags = ((1 << count) - 1);
      }
      else {
        curbytes = PACKETDATASIZE;
        packet.header.flags = ~(1 << count);
      }
      
      packet.header.size = curbytes;
      count++;
      nbytes -= curbytes;
      memcpy(packet.data, data, curbytes);
      data += curbytes;
      
      if (verbose)
        printf("transmitting: %d bytes left: %d\n",packet.header.size, nbytes);
      
      transmit_packet(tstate, &packet);

      if (param->polling)  poll(0, 0, param->polling);
    }
  }
}


/* *************************************************************************
 * void AudRecThread()
 *
 * ************************************************************************* */
void *AudRecThread(void *arg)
{
  int	read_size;
  int	rtn, wrtn;
  unsigned int error;
  TransceiveState *TxState;
  Parameters *param;
  
  param = (Parameters *) arg;

  TxState = (TransceiveState *) initialize_transmit (param, AUDIO);

  for ( ; ; ) {
    sem_wait (&AS);

    fprintf (stderr, "SendAudThread\n");
  
    /* Determine how much we can read */
    if (oti_ioctl(Audio_fd, FIONREAD, &read_size) < 0) {
      perror("record_service error: oti_ioctl FIONREAD");
      break;
    }

    oti_audio_get_record_error(Audio_fd, &error);
    if (error){
      error = 0;
      oti_audio_set_record_error(Audio_fd, &error);
    }

    if (read_size > 0) {
      rtn = oti_read(Audio_fd, aud_data, read_size);
      if (param->transmit) 
	BV_send_AudData (aud_data, rtn, TxState, param);
    }
    thrd_stop (&VR);  
  }
}


/* *************************************************************************
 * Function: int BV_recv_AudPacket(TransceiveState *RxState, PacketType *packet)
 *
 *
 * ************************************************************************* */
int BV_recv_AudPacket (TransceiveState *RxState, PacketType *packet, unsigned int *nbytes)
{
  unsigned int len;
  int fromlen;

  while ((pollfds[1].revents & POLLIN) == 0) {
    /* wait 100 msecs for packet */
    if (poll(&pollfds[1], 1, 100) < 0) {
      perror("poll failed");
      close(RxState->socket);
      free(RxState);
      return 0;
    }
  }

  /*  reset, so next time we will have to check at least once.   */
  pollfds[1].revents &= ~POLLIN;

  /*   len = read(RxState->socket, (char *) packet, sizeof(PacketType));  */

  fromlen = sizeof (RxState->name[RxState->session]);
  len = recvfrom(RxState->socket, (char *) packet, sizeof(PacketType), 0,
		 (struct sockaddr *) &RxState->name[RxState->session], 
		 (int *) &fromlen);

  nbytes = &len;
  return 1;
}


/* *************************************************************************
 * int BV_recv_AudData ()
 *
 *I
 * ************************************************************************* */
int BV_recv_AudData (TransceiveState *RxState, MsgType *message)
{
  PacketType packets[MAXPACKETS];	/* circular buffer of rec'vd packets */
  PacketType *packetlist[MAXPACKETS];	/* packets for this frame */
  PacketType *curpacket;
  int packetcnt = 0;
  int currentframe;
  int i;
  unsigned int nbytes;
  unsigned int currentsession;
  unsigned short myflags;
    
  /* .... receive video ....  */
  currentframe = 0;
  currentsession = 0;

  myflags = 0xffff;
  for (i = 0; i < MAXPACKETS; i++) packetlist[i] = NULL;

  curpacket = &packets[packetcnt++];
  if (packetcnt >= MAXPACKETS) packetcnt = 0;

  BV_recv_AudPacket(RxState, curpacket, &nbytes);

  if (currentsession != curpacket->header.session) {
    /*
     * if you join a conference in progress then the current frame and 
     * session will be grabbed from the first received packet received.
     */
    for (i = 0; i < MAXPACKETS; i++) 	packetlist[i] = NULL;
    myflags = 0xffff;
    currentframe = curpacket->header.frame;
    currentsession = curpacket->header.session;
  }

  if (curpacket->header.frame < currentframe) {
    /* if this packet is from an earlier frame disregard it   */
    if (verbose)
      fprintf(stderr, "received stale packet %d, current frame %d\n",
	      curpacket->header.frame, currentframe);
    return 1;
  }

  if (curpacket->header.frame > currentframe) {
    for (i = 0; i < MAXPACKETS; i++)	packetlist[i] = NULL;
    myflags = 0xffff;
    currentframe = curpacket->header.frame;
  }

  /* save the packet for decoding */
  packetlist[packetnum(curpacket->header.flags)] = curpacket;
  myflags &= curpacket->header.flags;

  if (myflags == 0)  create_message(message, packetlist);

  return 0; 
}


/* *************************************************************************
 * void *AudPlayThread(void *arg)
 *
 * ************************************************************************* */
void *AudPlayThread(void *arg)
{
  int rc, rtn;
  int  aud_amnt = 0;
  char *aud_ptr;
  char aud_data[BUFFER_SIZE];
  unsigned int data_size;
  unsigned int error;
  MsgType message;
  Parameters *param = (Parameters *) arg;
  TransceiveState *RxState;

  RxState = (TransceiveState *) initialize_receive (param, AUDIO);

  for ( ; ; ) {
    sem_wait (&AR);
    sem_post (&AS);

    fprintf (stderr, "RecvAudThread ");

    /* get data from netwrok */
    if (param->receive) {
      poll (0, 0, 5);
      RxState->session = 1;	/* choose remote host to receive */
      rc = BV_recv_AudData (RxState, &message);
      aud_amnt = message.header.size;
      aud_ptr = (char *) message.data;
    } 

    oti_audio_get_play_error(Audio_fd, &error);
    if (error){
      error = 0;
      oti_audio_set_play_error(Audio_fd, &error);
    }

    while (aud_amnt > 0) {
      /* write as much data as possible */
      rtn = oti_write(Audio_fd, aud_ptr, aud_amnt);
      
      if (rtn > 0) {
	aud_amnt -= rtn;
	aud_ptr += rtn;
      } else {
	break;
      }
    }
    thrd_stop (&VR);
  }
}  



/* *************************************************************************
 * BV_init_audio (audio_t *odo)
 *
 *
 * ************************************************************************* */
int BV_init_audio ()
{
  int   rc;
  char *aud_device = DEFAULT_DEVICE;
  char  audctl_device[80];
  char info[120];
  unsigned int value;

  Audio_hdr  hdr;
  Audio_info tmpinfo;

  /*
   * setup signal handlers
   */
  signal(SIGPOLL, BV_poll_handler);

  if (verbose) fprintf(stderr, "initializing audio device...\n");

  /*
   * This call must be made before any other
   */
  if ((rc = oti_audio_init(NULL, NULL)) < 0){
    fprintf(stderr,"oti_audio_init failed %d\n",rc);
    BV_exit(0);
  }
  
  /*
   * Open audio device and setup POLL signal
   */
  if ((Audio_fd = oti_open(aud_device, O_RDWR | O_NDELAY))<0){
    fprintf(stderr,"could not open osprey audio device\n");
    exit(0);
  }
  if (oti_ioctl(Audio_fd, I_SETSIG, (void*)(S_INPUT | S_OUTPUT | S_MSG)) < 0){
    perror("I_SETSIG ioctl");
    BV_exit(0);
  }

  /*
   * Open audio control device and setup POLL signal
   */
  get_audioctl_name(audctl_device, aud_device);
  if ((Audioctl_fd = oti_open(audctl_device, O_RDWR))<0){
    fprintf(stderr,"could not open audio ctl device\n");
    BV_exit(0);
  }    
  if (oti_ioctl(Audioctl_fd, I_SETSIG, (void*)(S_MSG)) < 0){
    perror("I_SETSIG ioctl");
    BV_exit(0);
  }

  /*
   * Set the desired input port and gains
   */
  oti_audio_set_record_port(Audio_fd, &port);
  oti_audio_set_record_gain(Audio_fd, &igain);
  oti_audio_set_monitor_gain(Audio_fd, &mgain);  
  oti_audio_set_play_gain(Audio_fd, &ogain);

  /*
   * Determine the desired format
   */
  switch (format)
  {
    case 0: hdr.encoding = AUDIO_ENCODING_ULAW;   break;
    case 1: hdr.encoding = AUDIO_ENCODING_ALAW;   break;
    case 2: hdr.encoding = AUDIO_ENCODING_LINEAR; break;
    case 3: hdr.encoding = AUDIO_ENCODING_LINEAR; break;
    case 4: hdr.encoding = AUDIO_ENCODING_G722;   break;
    case 5: hdr.encoding = AUDIO_ENCODING_G728;   break;
    default: hdr.encoding = AUDIO_ENCODING_ULAW;
  }

  switch (rate)
  {
    case 8000:  hdr.sample_rate =  8000; break;
    case 11025: hdr.sample_rate = 11025; break;
    case 16000: hdr.sample_rate = 16000; break;
    case 22050: hdr.sample_rate = 22050; break;
    default:    hdr.sample_rate =  8000; 
  }

  switch (channels)
  {
    case 1:  hdr.channels = 1; break;
    case 2:  hdr.channels = 2; break;
    default: hdr.channels = 1; break;
  }

  switch(hdr.encoding)
  {
    case AUDIO_ENCODING_ULAW:
    case AUDIO_ENCODING_ALAW:    
    {
      hdr.samples_per_unit = 1;
      hdr.bytes_per_unit = 1;
      break;
    }
    case AUDIO_ENCODING_LINEAR:
    {
      hdr.samples_per_unit = 1;
      if (format == 2)
        hdr.bytes_per_unit = 1; /* 8 bit PCM */
      else
        hdr.bytes_per_unit = 2; /* 16 bit PCM */
      break;
    }
    case AUDIO_ENCODING_G722:
    {
      /* G722 fixed rate and channels */
      hdr.sample_rate = 8000;
      hdr.channels    = 1;
      hdr.samples_per_unit = 1;
      hdr.bytes_per_unit = 1;
      break;
    }
    case AUDIO_ENCODING_G728:
    {
      /* G728 fixed rate and channels */
      hdr.sample_rate = 1600;
      hdr.channels    = 1;
      hdr.samples_per_unit = 8;
      hdr.bytes_per_unit = 10;
      break;
    }
  }
  
  hdr.data_size = AUDIO_UNKNOWN_SIZE;
  hdr.endian = AUDIO_ENDIAN_BIG;

  /*
   * Pause the input and flush the read queue
   */
  (void) oti_audio_pause_record(Audio_fd);
  (void) oti_audio_flush_record(Audio_fd);

  /*
   * Set the desired buffer size
   */
  AUDIO_INITINFO(&tmpinfo);
  tmpinfo.record.buffer_size = buffer_size;
  tmpinfo.play.buffer_size = buffer_size;
  oti_audio_setinfo(Audio_fd, &tmpinfo);

  /*
   * Set the format
   */
  oti_audio_set_play_config(Audio_fd, &hdr);
  value = 0;
  oti_audio_set_play_eof(Audio_fd, &value);
  oti_audio_set_record_config(Audio_fd, &hdr);
  
  /*
   * Pause the input and flush the read queue
   */
  (void) oti_audio_pause_record(Audio_fd);
  
  AUDIO_INITINFO(&tmpinfo);
  tmpinfo.record.pause = 0;
  tmpinfo.record.samples = 0;
  tmpinfo.record.error = 0;
  tmpinfo.play.pause = 0;
  tmpinfo.play.samples = 0;
  tmpinfo.play.error = 0;
  oti_audio_setinfo(Audio_fd, &tmpinfo);

  (void) oti_audio_flush_record(Audio_fd);

  return 0;
}

