
/* *************************************************************************
 *  File : MTvodo.c
 * -------------------------------------------------------------------------
 *  All rights are reserved by Bo-kyun Na, IntellActive Inc.
 * -------------------------------------------------------------------------
 * This example reads an image from MMACo1k, transmits compressed video over
 * the local network, and optionally displays the image in a Xwindow.
 *
 * ************************************************************************* */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include <sys/audioio.h>

#include <multimedia/libaudio.h>
#include <multimedia/audio_device.h>
#include <multimedia/audio_encode.h>

#include "params.h"
#include "network.h"
#include "transceive.h"

#define BUFFER_SIZE (1024*4)
#define _POSIX_C_SOURCE 199506L

pthread_mutex_t utex_lock = PTHREAD_MUTEX_INITIALIZER;

int count = 0;
sem_t VS, VR, AS, AR;

extern int verbose;
extern void create_message_h261 (MsgType *msg, PacketType *packetlist[]);
extern int input_fd, output_fd;

void *Thread_TxVideo (void *);
void *Thread_RxVideo (void *);
extern void *Thread_TxAudio (void *);
extern void *Thread_RxAudio (void *);


/* *************************************************************************
 * thrd_stop (sem_t *sem)
 *
 * ************************************************************************* */
void thrd_stop (sem_t *sem)
{
  pthread_mutex_lock (&utex_lock);
  (count) ? sem_post (sem), count = 0 : count++;
  pthread_mutex_unlock (&utex_lock);
}


/* *************************************************************************
 * usage ()
 *
 *
 * ************************************************************************* */
void usage (char *pgm)
{
  fprintf (stderr, 
"usage: $s [options] \n \
Legal options:                                       Defaults\n \
<-i file_name> : audio input file name               0\n \
<-o file_name> : audio output file name              0\n \
", pgm);
}


/* *************************************************************************
 * void *Thread_TxVideo (void *arg)
 *
 *
 * ************************************************************************* */
void *Thread_TxVideo (void *arg)
{
  int scalex = 1, scaley = 1;
  XilImage image, o1k_image;
  XilDataType datatype;
  XilImage scaled_image;
  XilCis cis;
  XilCis cis2;    
  Xil_unsigned8 *data[2];
  unsigned int di  = 0;
  unsigned int dip = 0;    
    
  Parameters *param = (Parameters *) arg;
  TransceiveState *tstate;
  WindowState *wstate;

  unsigned int width, height, nbands;
  int frames;
  double t;
  int nframes[2] = { 0, 0};
  int nbytes[2];
  int multiple_frames = 0;
  const char *video_format;
  int bytes_for_first_frame;
  FILE *df;
  int first_frame = 1;
  int format;

  if (param->transmit)
    tstate = (TransceiveState *) initialize_transmit(param, VIDEO);

  {
    XilDevice device;

    if (!(device = xil_device_create(xil_state, "MMACo1k"))) {
      fprintf(stderr, "Unable to create a device object\n");
      xil_close(xil_state);
      exit(1);
    }

    xil_device_set_value(device, "DEVICE_NAME", (void *) param->devname);
    if (!(image = xil_create_from_device(xil_state, "MMACo1k", device))) {
      fprintf(stderr, "ERROR: failed to open MMACo1k device\n");
      xil_close(xil_state);
      exit(1);
    }
    xil_device_destroy(device);
  }
  o1k_image = image;

  xil_set_device_attribute(image, "PORT_V", (void *) param->whichinput);

  xil_get_device_attribute(o1k_image, "FORMAT_V", (void **) &format);
  if (format == 0) {
    fprintf(stderr, "Unknown video format, exiting.\n");
    xil_close(xil_state);
    exit(1);
  }

  xil_set_device_attribute(image, "IMAGE_SKIP", (void *) param->image_skip);
  xil_set_device_attribute(image, "MAX_BUFFERS", (void *) param->max_buffer);
  xil_get_device_attribute(image, "FORMAT", (void **) &video_format);

  xil_get_info(image, &width, &height, &nbands, &datatype);

    /*
     * If the user has not changed the default to something else, reset the
     * default capture width and height
     */
  if (param->width == -1)
    param->width = width;
  if (param->height == -1)
    param->height = height;

    /* check for width or height too big */
  if ((param->width > width) || (param->height > height)) {
    fprintf(stderr, "ERROR: Maximum width & height are %d & %d\n", 
	    width, height);
    xil_close(xil_state);
    exit(1);
  }

  /*
   * if user specifies '-s 0' for no scaling, set local 'scalefactor' to 1.0
   */
  scalex = (param->scalex ? param->scalex : 1);
  scaley = (param->scaley ? param->scaley : 1);
    
  if ((param->width < width) || (param->height < height)) {
    image = xil_create_child(o1k_image, (width - param->width) / 2,
			     (height - param->height) / 2,
			     param->width, param->height, 0, nbands);
    if (image == NULL) {
      fprintf(stderr, "ERROR: failed to open child image on MMACo1k\n");
      xil_close(xil_state);
      exit(1);
    }
    xil_get_info(image, &width, &height, &nbands, &datatype);
  }

  /*
   * H261 CIF is defined as 352x288 not 320x240. We set the width
   * and height here and create a scaled_image which will be used
   * in the capture/scale8nearest/compress_H261 molecule
   */
  if (param->do_cis == H261) {
    xil_set_device_attribute(o1k_image, "H261_BIT_RATE", 
			     (void *) param->h261_bitrate);
    xil_set_device_attribute(o1k_image, "H261_MAX_QUANT", 
			     (void *) param->h261_maxquant);
    if (scalex == 2){
      width = 352;
      height = 288;
    } else if (scalex == 4) {
      width = 352/2;
      height = 288/2;
    } else {
      fprintf(stderr,"Invalid scale factor for H261 Compression: only 2 and 4 legal\n");
      xil_close(xil_state);
      exit(1);
    }
    if (verbose)
      printf("h261: scaled_image w: %d h: %d b: %d\n", width, height, nbands);
    scaled_image = xil_create(xil_state, width, height, nbands, XIL_BYTE);
  } else {
    if (scalex || scaley) {
      width = (int) ((float) width / (float) scalex + 0.5);
      height = (int) ((float) height / (float) scaley + 0.5);
      scaled_image = xil_create(xil_state, width, height, nbands, XIL_BYTE);
    }
  }
  if (param->preview)
    wstate = initialize_window(param, xil_state);

  cis  = xil_cis_create(xil_state, param->cis_type);

    /*
     * set the buffers to make sure that you don't keep information you do
     * not need, but H261 currently needs to keep more frames around do
     * to XIL CIS seeking problems.
     */
    
  if (param->do_cis == H261) {

      /*
       * H261 needs to keep more than one frame in the CIS
       * because of the non-byte aligned nature of H261. XIL
       * has chosen to mark the beginning of a H261 frame by
       * the bytes that encompass the H261 PSC. These bytes,
       * however, may contain bits from the previous frame.
       * Thus for a given H261 frame, you need at least that
       * frame and the next frame in the CIS.
       */
      
    xil_cis_set_keep_frames(cis, 3);
    xil_cis_set_max_frames(cis, 3);
      
      /*
       * The XIL H261 CIS is not seekable. This causes a problem
       * if one tries to both read data from a CIS (to transmit)
       * and also display from the CIS (for local preview). The
       * way this is normally done is to retrieve data from the
       * CIS (with a get_bits call) and then seek back a frame
       * and the call decompress on this frame to display it.
       * With the seek problem this is not possible. The solution
       * to the problem is to create a second CIS. When data is
       * retrieved from the CIS it is added to the second CIS as
       * well as being transmitted. Decompression is out of the
       * second CIS.
       */
      
    cis2 = xil_cis_create(xil_state, param->cis_type);
      
    xil_cis_set_keep_frames(cis2, 3);
    xil_cis_set_max_frames(cis2, 3);
      
  } 
  else {
    xil_cis_set_keep_frames(cis, 1);
    xil_cis_set_max_frames(cis, 1);
  }

    /*
     * We need to be able to seek backwards to both get-bits, and decompress
     */
  if (param->do_cis == JPEG) {
	/* this attribute must be TRUE for acceleration */
    xil_cis_set_attribute(cis, "ENCODE_411_INTERLEAVED", (void *) TRUE);
  }

  if (param->datafile)
    df = fopen(param->datafile, "w");
  else
    df = (FILE *) NULL;

  for (;;) {
    sem_wait (&VS);

    fprintf (stderr, "SendVidThread ");

      /* check for X events like a `quit' from the pulldown menu */
    if (param->preview)
      check_events(wstate);
      
    if (param->do_cis == H261){
      /* This is the special scale molecule to handle the
       * 352x288 CIF H261 specification.                       */
      if (format == 2) /* NTSC */
	xil_scale(image, scaled_image, "nearest", width/640.0, height/480.0);
      else /* PAL */
	xil_scale(image, scaled_image, "nearest", width/768.0, height/576.0);
    } else if (scalex || scaley) {
      xil_scale(image, scaled_image, "nearest", 1.0 / (float) scalex, 
		1.0 / (float) scaley);
    }
    else {
      scaled_image = image;
    }
      
    xil_compress(scaled_image, cis);
    xil_cis_sync(cis);
    xil_toss(scaled_image);
      
    if (xil_cis_has_frame(cis)) {
      /* If we're not in the midst of handling multiple frames, retrieve
       * as many as we can via xil_cis_get_bits_ptr().                   */

      if (!multiple_frames) {
          /*
           * There are many options one can use to packetize H261. Some
           * methods are easier if the decoder can take zero stuffing.
           * H261 can not handle zero stuffing but it can handle macro
           * block stuffing. More documentation on this is in the
           * receiver program. In any rate, an H261 frame in XIL is
           * completed by possible bits in the following frame. These
           * bits need to be transmitted as part of the frame in case
           * the next frame gets dropped by the network. On the
           * receiver end, a valid bitstream can be constructed even
           * if a frame is dropped.
           *
           * In sum, the above is the reason we are double buffering the
           * data retrieved from the H261 CIS. This way, we can send to
           * the transmit routine the current frame and the last bits for
           * the frame which are actually in the first bits of the next
           * frame.
           */
           
	data[di] = (void *) xil_cis_get_bits_ptr (cis, &nbytes[di], &nframes[di]);
          
	if (param->do_cis == H261) {

	  /*
	   * Since an XIL H261 frame is non-byte aligned and because XIL
	   * marks the beginning of a frame by the bytes that make up the 
	   * PSC of the H261 frame, the data retrieved from get bits will
	   * not be a complete frame. Thus the 0 parameter in the call to
	   * xil_cis_put_bits below
	   */
            
	  xil_cis_put_bits(cis2, nbytes[di], 0, data[di]);
	}
          
	if (df)
	  fwrite(data[di], 1, nbytes[di], df);
	
	/* Packetization and sending of video data */
	if (param->transmit) 
	  Vid_TxData(data[di], nbytes[di], width, height, tstate, param);
          
	if(param->do_cis == H261) {/* double buffer H261 CIS frame retrieval */
	  dip = di;
	  di = (di+1)&0x1;
	}
          
	first_frame = 0;
      }
        
      if (param->preview) {
	if (param->do_cis == H261) {
            /*
             * There must be at least two frames in an XIL H261 CIS
             * in order to decompress from it.
             */
	  if (xil_cis_number_of_frames (cis2) > 2)
              display(wstate, cis2);
	} 
	else {
            /*
             * xil_cis_get_bits() moved the CIS read frame forward by
             * nframes. To decompress these, seek back nframes. If we're
             * in the midst of handling multiple frames, no seek is
             * needed.
             */
	  if (!multiple_frames)
	    xil_cis_seek(cis, -nframes[dip], 1);
	  display(wstate, cis);
	}
      }
    }
      
    if (nframes[dip] > 0) {
        /*
         * If we're in the midst of handling multiple frames, decrement
         * the count.  Else, set multiple_frames if nframes is more than
         * 1.
         */
      if (multiple_frames)
	multiple_frames--;
      else
	multiple_frames = nframes[dip] - 1;
    }

   thrd_stop (&AR); 
   /*   thrd_stop (&AS); */
  }
}


/* ********************************************************************
 * Thread : void *Thread_RxVideo(void *arg)
 *
 *
 * ******************************************************************** */
void *Thread_RxVideo (void *arg)
{
  Parameters *params = (Parameters *) arg;
  TransceiveState *receiver;
  WindowState *window;
  PacketType packets[MAXPACKETS];	/* circular buffer of rec'vd packets */
  PacketType *packetlist[MAXPACKETS];	/* packets for this frame */
  PacketType *curpacket;
  MsgType message;
  int packetcnt = 0;
  int currentframe;
  unsigned int currentsession;
  int lastw, lasth;
  int i;
  unsigned short myflags;
  XilCis cis;
    
  long display_frames = 0;

  /* .... initialize everything ....  */
  receiver = initialize_receive(params, VIDEO);

  receiver->session = 1;	/* choose remote host to receive */
  /*  (char *) receiver->name[receiver->session].sin_addr = 
      (char *) params->hostname[receiver->session]; 
      printf ("remotehost = %s\n",receiver->name[receiver->session]);
   */
  cis = xil_cis_create(xil_state, params->cis_type);

  if (params->do_cis == H261) 
    xil_cis_set_max_frames(cis, 4);
  else
    xil_cis_set_max_frames(cis, 1);    

  if ((window = initialize_window(params, xil_state)) == NULL) {
    fprintf(stderr, "ERROR: Unable to create window. ");
    xil_close(xil_state);
    exit(-1);
  }

  /* .... receive video ....      */
  currentframe = 0;
  currentsession = 0;
  lastw = params->width;
  lasth = params->height;

  /* H261 CIF is defined as 352x288 not 320x240. We set the width
   * and height here.
   */
  if (params->do_cis == H261) {
    lastw = 352;
    lasth = 288;
  }
    
  /*
   * If JPEG, compress and decompress a dummy frame to load all the
   * necessary tables for the decompressor.  This allows us to decompress
   * received frames created with COMPRESSED_DATA_FORMAT set to
   * ABBREVIATED_FORMAT even if we miss the first frame. NOTE: this
   * technique assumes that the compressed bitstream was created using the
   * default JPEG CIS attributes.
   */
  if (params->do_cis == JPEG) {
    XilImage dummy = xil_create(xil_state, lastw, lasth, 3, XIL_BYTE);

    xil_compress(dummy, cis);
    xil_decompress(cis, dummy);
    xil_sync(dummy);
    xil_destroy(dummy);
  }

    /*
     * Because of the non-byte aligned nature of H261 and because
     * the XIL H261 CIS can not tolerate zero stuffing, the first
     * frame added to an H261 CIS must begin with a perfect PSC.
     * Below we add a dummy blank (a black H261 frame) that starts
     * with a perfect PSC. Since the XIL H261 CIS needs possibly at
     * least two frames in order to decompress a given frame, we
     * add the blank frame twice.
     */
     
  if (params->do_cis == H261) {
    xil_cis_put_bits(cis, sizeof(blank_261_frame_cif), 0, blank_261_frame_cif);
  }
    
  myflags = 0xffff;
  for (i = 0; i < MAXPACKETS; i++)
    packetlist[i] = NULL;

  for (;;) {
    sem_wait (&VR);
    sem_post (&VS); 

    fprintf (stderr, "RecvVidThread ");

    curpacket = &packets[packetcnt++];
    if (packetcnt >= MAXPACKETS)
      packetcnt = 0;

    /* Get video packet if available.  It no video then the window state
     * is passed in so we can at least check for X events.  (If the user
     * pulled down a `quit' menu.)
     */
    receive_packet(receiver, curpacket);

    if (currentsession != curpacket->header.session) {
      /* if you join a conference in progress then the current frame and 
       * session will be grabbed from the first received packet received.  */
      if (verbose) {
	fprintf(stderr, "changing session from %d to %d\n",
		currentsession, curpacket->header.session);
      }
      for (i = 0; i < MAXPACKETS; i++)
	packetlist[i] = NULL;
      myflags = 0xffff;
      currentframe = curpacket->header.frame;
      currentsession = curpacket->header.session;
    }

    /* if this is the first packet of a frame, check for a resize */
    if ((curpacket->header.flags & 1) == 0) {
      /* the size of each frame is sent on the first packet of each frame */
      if ((curpacket->header.x != lastw) || (curpacket->header.y != lasth)) {
	lastw = curpacket->header.x;
	lasth = curpacket->header.y;

	if (verbose) printf("resizing video %dx%d\n",lastw, lasth);

	resize(window, lastw, lasth);
	xil_cis_reset(cis);

	/* see comments above about dummy compress/decompress */
	if (params->do_cis == JPEG) {
	  XilImage dummy = xil_create(xil_state, lastw, lasth, 3, XIL_BYTE);
	  xil_compress(dummy, cis);
	  xil_decompress(cis, dummy);
	  xil_sync(dummy);
	  xil_destroy(dummy);
	}
            
	/* see comments above */
	if (params->do_cis == H261) {
	  if (lastw == 352)
	    xil_cis_put_bits(cis, sizeof(blank_261_frame_cif), 0, 
			     blank_261_frame_cif);
	  else
	    xil_cis_put_bits(cis, sizeof(blank_261_frame_qcif), 0, 
			     blank_261_frame_qcif);                  
	}
	if (params->do_cis == H261)
	  xil_cis_set_max_frames(cis, 4);
	else
	  xil_cis_set_max_frames(cis, 1);
      }
    }

    if (curpacket->header.frame < currentframe) {
      /* if this packet is from an earlier frame disregard it    */
      if (verbose)
	fprintf(stderr, "received stale packet %d, current frame %d\n",
		curpacket->header.frame, currentframe);
      continue;
    }
    if (curpacket->header.frame > currentframe) {

      /* If the frame number increased then we probably missed the last
       * packet of the previous frame. We display what we can of the
       * previous frame if '-b' option was selected, otherwise we drop
       * the partial frame.                                              */
      if (myflags != 0) {
	if (verbose)
	  fprintf(stderr, "dropping %d due to getting packet %d (flags %x)\n",
		  currentframe, curpacket->header.frame, myflags);
      }

      for (i = 0; i < MAXPACKETS; i++)
	packetlist[i] = NULL;
      myflags = 0xffff;
      currentframe = curpacket->header.frame;
    }

    /*
     * save the packet for decoding
     */
        
    packetlist[packetnum(curpacket->header.flags)] = curpacket;
    myflags &= curpacket->header.flags;

    if (myflags == 0) {
      if (params->do_cis == H261){
	/* H261 is very sensitive to packet lost. This routine fixes
	 * any problems with a lost packet                              */
	create_message_h261 (&message, packetlist);
      } else
	create_message(&message, packetlist);

      xil_cis_put_bits(cis, message.header.size, 0, message.data); 

      display_frames++;
      if (params->do_cis == H261){
	/* we could check to make sure there is at least two frames 
	 * in the XIL H261 CIS at this point but since we've added
	 * a dummy blank frame to start with and we only call
	 * display after we've added a frame, we don't need to check.   */
	display(window, cis);
      } else 
	display(window, cis);
    }
    thrd_stop (&AR); 
    /*    thrd_stop (&AS); */
  }
}


/* ***************************************************************
 * Function: void main (int argc, char **argv) 
 *
 *
 * *************************************************************** */
void main (int argc, char **argv)
{
  int   i;
  char aud_fname[80];
  Parameters *Tparam,
             *Rparam;

  pthread_t tid_main,
            tid_VidTx,
            tid_VidRx;

  pthread_t tid_AudRec, 
            tid_AudPlay;

  pthread_attr_t attr;

  SEM_INIT (&VS, NULL, 0);
  SEM_INIT (&VR, NULL, 0);
  SEM_INIT (&AS, NULL, 0);
  SEM_INIT (&AR, NULL, 0);

  PTHREAD_ATTR_INIT (&attr);
  pthread_attr_setscope (&attr, PTHREAD_SCOPE_SYSTEM);


  for (i=1; i<argc; i++){
    switch (argv[i][1])
      {
      case 'i':
	strcpy(aud_fname, argv[++i]);
	if (audio_isaudiofile(aud_fname)) {
	  input_fd = open(aud_fname, O_RDONLY, 0666);
	  if (input_fd < 0)
	    fprintf(stderr,"could not open %s\n", aud_fname);
	} 
	else {
	  fprintf(stderr,"file %s is not an audio file\n", aud_fname);
	}
	break;
      case 'o':
	strcpy(aud_fname, argv[++i]);
	output_fd = open(aud_fname, O_WRONLY | O_CREAT | O_TRUNC, 0666);
	if (output_fd < 0)
	  fprintf(stderr,"could not create %s\n", aud_fname);
	break;
      case '?':
      default:
        usage(argv[0]);
        exit(0);
      }
  }

  /* Initialization of video devices */
  Tparam = (Parameters *) init_tvparam(argc, argv);
  Rparam = (Parameters *) init_rvparam(argc, argv);
  
  Tparam->data_via_file = Rparam->data_via_file = 0;

  xil_state = xil_open();
  if (xil_state == NULL) {
    fprintf(stderr, "unable to open xil library\n");
    exit(-1);
  }

  sem_post (&VR);

  /* Creation of send/receive video threads */
  sleep (1);
  PTHREAD_CREATE (&tid_VidRx, &attr, Thread_RxVideo, (void *) Rparam);

  sleep (1);
  PTHREAD_CREATE (&tid_VidTx, &attr, Thread_TxVideo, (void *) Tparam);

  /* Initialization of audio devices */
  if (BV_init_audio ()) {
    fprintf (stderr, "Error: BV_init_audio\n");
    exit (0);
  }

  sleep (1);
  PTHREAD_CREATE (&tid_AudPlay, &attr, AudPlayThread, (void *) Rparam);

  sleep (1);
  PTHREAD_CREATE (&tid_AudRec, &attr, AudRecThread, (void *) Tparam);

  pthread_exit ((void *) NULL);
}
